use std::io::{stdin, stdout, Write};

use anyhow::{bail, Result};
use termion::{color, cursor::HideCursor, event::Key, input::TermRead, raw::IntoRawMode};

use crate::{
	board::{Board, PlayerAction},
	util::{Player, Symb},
};

use super::{MaximizerAgent, MinimizerAgent};

struct SymbolSelector {
	symbols: Vec<char>,
	cursor: usize,
}

impl SymbolSelector {
	fn new(symbols: Vec<char>) -> Self {
		Self { symbols, cursor: 0 }
	}

	fn current(&self) -> char {
		self.symbols[self.cursor]
	}

	fn check(&mut self, board: &Board) {
		while board.contains(Symb::from_char(self.current()).unwrap()) {
			if self.cursor == 0 {
				self.cursor = self.symbols.len() - 1;
			} else {
				self.cursor -= 1;
			}
		}
	}

	fn up(&mut self, board: &Board) {
		if self.cursor == 0 {
			self.cursor = self.symbols.len() - 1;
		} else {
			self.cursor -= 1;
		}

		while board.contains(Symb::from_char(self.current()).unwrap()) {
			if self.cursor == 0 {
				self.cursor = self.symbols.len() - 1;
			} else {
				self.cursor -= 1;
			}
		}
	}

	fn down(&mut self, board: &Board) {
		if self.cursor == self.symbols.len() - 1 {
			self.cursor = 0;
		} else {
			self.cursor += 1;
		}

		while board.contains(Symb::from_char(self.current()).unwrap()) {
			if self.cursor == self.symbols.len() - 1 {
				self.cursor = 0;
			} else {
				self.cursor += 1;
			}
		}
	}
}

pub struct PlayerAgent {
	player: Player,
	cursor: usize,
	symbol_selector: SymbolSelector,
}

impl PlayerAgent {
	pub fn new(player: Player) -> Self {
		Self {
			player,
			cursor: 0,
			symbol_selector: SymbolSelector::new(vec![
				'1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '+', '-', '×', '÷',
			]),
		}
	}

	fn step(&mut self, board: &Board, minimize: bool) -> Result<PlayerAction> {
		let stdout = HideCursor::from(stdout().into_raw_mode().unwrap());
		let mut stdout = stdout.lock();
		let cursor_max = board.size() - 1;

		self.symbol_selector.check(board);

		// Ask for input until we get a valid move
		loop {
			print!(
				"\r{}{}{}   ╙{}{}{}{}{}╜",
				// Goal
				color::Fg(self.player.color()),
				if minimize { "Min" } else { "Max" },
				color::Fg(color::Reset),
				// Cursor
				" ".repeat(self.cursor),
				color::Fg(self.player.color()),
				if board.is_done() {
					' '
				} else {
					self.symbol_selector.current()
				},
				color::Fg(color::Reset),
				" ".repeat(cursor_max - self.cursor),
			);
			stdout.flush()?;

			// Player turn
			let stdin = stdin();
			for c in stdin.keys() {
				match c.unwrap() {
					Key::Char('q') => bail!("player ended game"),
					Key::Right => {
						self.cursor = cursor_max.min(self.cursor + 1);
						break;
					}
					Key::Left => {
						if self.cursor != 0 {
							self.cursor -= 1;
						}
						break;
					}
					Key::Up => {
						self.symbol_selector.up(board);
						break;
					}
					Key::Down => {
						self.symbol_selector.down(board);
						break;
					}
					Key::Char('\n') => {
						let symb = Symb::from_char(self.symbol_selector.current());
						if let Some(symb) = symb {
							let action = PlayerAction {
								symb,
								pos: self.cursor,
							};
							if board.can_play(&action) {
								return Ok(action);
							}
						}
						break;
					}
					Key::Char(c) => {
						let symb = Symb::from_char(c);
						if let Some(symb) = symb {
							let action = PlayerAction {
								symb,
								pos: self.cursor,
							};
							if board.can_play(&action) {
								return Ok(action);
							}
						}
						break;
					}
					_ => {}
				};
			}
		}
	}
}

impl MinimizerAgent for PlayerAgent {
	fn step_min(&mut self, board: &Board) -> Result<PlayerAction> {
		self.step(board, true)
	}
}

impl MaximizerAgent for PlayerAgent {
	fn step_max(&mut self, board: &Board) -> Result<PlayerAction> {
		self.step(board, false)
	}
}
